--------------------------------------------------------------------------------
-- Stand Alone Uplink: FPGA Design
-- LIGO document: D060288
-- SubModule RX_Trig
-- Created   9/25/2006 3:49:25 PM
--
-- RX_Trig generates the signals to mark the data time stamp, the receive
-- window and the state machine triggers. The receive window is the time
-- duration in which valid data belonging to a given time stamp can be
-- successfully received. Currently, the receive window ends 4.29us ahead
-- of the data time stamp. The beginning of the window for the next processing
-- cycle start immediately afterwards. This means data has to arrive 4.29us
-- before it is needed.
--
-- Look in TrigQueue for a description of the data triggering algorithm.
--
-- The state machine is triggerd 1.91us (2^-19s) ahead of the data time stamp.
-- Since the data rate from the eithernet is about a factor of 2.15 slower
-- than the processing speed of the microcode state machine, this together
-- with the location of the receive window guarantees that incoming data
-- fully resides in the queue before it is needed.
--
-- Parameters:
-- CLK: clock
-- GPS: GPS seconds
-- CNT: GPS fractional seconds in units of 2^-26
-- TRIG: trigger time of last trigger in bundle
-- MASK: Determines the trigger repetition rate
-- ENDM: Determines the bundling of trigger
-- GO: Trigger (applied for one cycle)
-- FIRST: Marks the first trigger in a bundle (applied for one cycle)
-- LAST: Marks the last trigger in a bundle (applied for one cycle)
-- FLUSH: Flush will be applied to indicate that the triggering has been
--        disabled. It should be used to flush buffers and reset the state
--        machines which process the triggered data (applied for one cycle)
-- RECEIVE: Indicates the beginning of the receive window
-- TSTAMP: Provides the valid time stamp for the receive window
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity RX_Trig is port
   (
     CLK         : in    std_logic;
     GPS         : in    std_logic_vector(31 downto 0);
     CNT         : in    std_logic_vector(25 downto 0);
     TRIG        : in    std_logic_vector(31 downto 0);
     MASK        : in    std_logic_vector(31 downto 0);
     ENDM        : in    std_logic_vector(31 downto 0);
     GO          : out   std_logic;
     FIRST       : out   std_logic;
     LAST        : out   std_logic;
     FLUSH       : out   std_logic;
     RECEIVE     : out   std_logic;
     TSTAMP      : out   std_logic_vector(63 downto 0)
   );
end RX_Trig;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of RX_Trig is

-- Constant Declarations
   constant go_offset  : unsigned (25 downto 0) := B"00" & X"000080";
   constant rcv_offset : unsigned (25 downto 0) := B"00" & X"000120";

-- Component Declaration
component TrigQueue port
   (
     CLK         : in    std_logic;
     ENABLE      : in    std_logic;
     CNT         : in    std_logic_vector(25 downto 0);
     TRIG        : in    std_logic_vector(25 downto 0);
     MASK        : in    std_logic_vector(25 downto 0);
     ENDM        : in    std_logic_vector(25 downto 0);
     GO          : out   std_logic;
     FIRST       : out   std_logic;
     LAST        : out   std_logic;
     FLUSH       : out   std_logic
   );
end component TrigQueue;

-- Signal Declarations
   signal tstamp_go : std_logic;
   signal tstamp_first : std_logic;
   signal gps_s : unsigned (31 downto 0);
   signal gps_f : unsigned (25 downto 0);
   signal proc_cycle : unsigned (26 downto 0);
   signal tstamp_ofs : unsigned (57 downto 0);
   signal fsm_go : std_logic;
   signal fsm_first : std_logic;
   signal fsm_last : std_logic;
   signal fsm_flush : std_logic;
   signal rcv_go : std_logic;
   signal rcv_first : std_logic;

   signal trig_go  : std_logic_vector (25 downto 0);
   signal trig_rcv : std_logic_vector (25 downto 0);

begin
   -- compute state machine go and receiver window difference
   trig_go  <= std_logic_vector (unsigned (TRIG (31 downto 6)) - go_offset);
   trig_rcv <= std_logic_vector (unsigned (TRIG (31 downto 6)) - rcv_offset);

   -- time stamp trigger
   time_stamp: TrigQueue port map
   (
      CLK => CLK,
      ENABLE => '1',
      CNT => CNT,
      TRIG => TRIG (31 downto 6),
      MASK => MASK (31 downto 6),
      ENDM => ENDM (31 downto 6),
      GO => tstamp_go,
      FIRST => tstamp_first,
      LAST => open,
      FLUSH => open
   );
   -- mark time stamp
   process (CLK, tstamp_go, tstamp_go) is
   begin
      if rising_edge (CLK) then
         if (tstamp_go = '1') and (tstamp_first = '1') then
            gps_s <= unsigned (gps);
            gps_f <= unsigned (cnt);
         end if;
      end if;
   end process;
   -- processing cycle length is one more than the end mask
   -- use a pipeline to relax the timing constraints
   process (CLK, ENDM) is
   begin
      if rising_edge (CLK) then
         proc_cycle <= unsigned (ENDM (31 downto 6)) + (B"000" & X"000001");
      end if;
   end process;
   -- add 2 * processing cycles to the time stamp
   process (CLK, tstamp_ofs) is
   begin
      if rising_edge (CLK) then
         tstamp_ofs <= (GPS_s & GPS_f) + (B"00" & X"0000000" & proc_cycle & B"0");
      end if;
   end process;
   -- now synchronize with the receiving window
   -- we are now a little more than one processing cycle ahead
   process (CLK, rcv_go, rcv_first, tstamp_ofs) is
   begin
      if rising_edge (CLK) then
         if (rcv_go = '1') and (rcv_first = '1') then
            TSTAMP (63 downto 6) <= std_logic_vector (tstamp_ofs);
         end if;
      end if;
   end process;
   TSTAMP (5 downto 0) <= (others => '0');

   -- state machine go
   go_fsm: TrigQueue port map
   (
      CLK => CLK,
      ENABLE => TRIG(0),
      CNT => CNT,
      TRIG => trig_go,
      MASK => MASK (31 downto 6),
      ENDM => ENDM (31 downto 6),
      GO => fsm_go,
      FIRST => fsm_first,
      LAST => fsm_last,
      FLUSH => fsm_flush
   );
   GO <= fsm_go;
   FIRST <= fsm_first;
   LAST <= fsm_last;
   FLUSH <= fsm_flush;

   -- receive window
   rcv_window: TrigQueue port map
   (
      CLK => CLK,
      ENABLE => '1',
      CNT => CNT,
      TRIG => trig_rcv,
      MASK => MASK (31 downto 6),
      ENDM => ENDM (31 downto 6),
      GO => rcv_go,
      FIRST => rcv_first,
      LAST => open,
      FLUSH => open
   );
   RECEIVE <= rcv_go and rcv_first;

end Structure;
--------------------------------------------------------------------------------
