-------------------------------`-------------------------------------------------
-- Stand Alone Uplink: FPGA Design
-- LIGO document: D060288
-- SubModule RX_arbiter_ABCD
-- Created   9/25/2006 5:02:25 PM
--
-- The RX_arbiter receives ethernet packages and dispatches them into one
-- of five queues. A, B, C and D queues are for isosynchronous transfers,
-- whereas the LP queue is for lower priority traffic. These queues cross
-- the clock boundary between the ethernet MAC and the user logic.
--
-- The LP is a simple first in first out queue 16 bits wide. It uses an
-- additional bit to mark the last word of a package. The LSBit of the
-- status word indicates if there is somethiing in this queue. The full
-- ethernet frame will be trasnferred to the LP queue. The maximum size of
-- this FIFO is 4kB.
--
-- The isosynchronous queues are implemented as double buffers. Each buffer
-- is fully dual ported and is written with data from the eithernet MAC. It is
-- read by the user logic. The double buffering makes sure that new data
-- can be received whilst the old one is still processeced. Only the payload
-- is transferred to this buffer. All other frame paramters such as time
-- stamp and packet length are checked beforehand and only valid payloads
-- are forwarded. The maximum size of one of the individual buffers is 4kB
-- or 1024 DAC values.
--
-- The RX-arbiter requires 3(LP) + 4(A-D) * 2(double buffer) * 2(size) = 19
-- block RAMs.
--
-- TODO: Currently CRC errors in an ethernet frame are ignored.
--
-- Parameters:
-- EMAC signals:
-- rx_clk: GIG EMAC RX clock, ie., 62.5 MHz
-- rx_data: GIG EMAC RX data word
-- rx_data_valid: GIG EMAC data valid bit
-- rx_data_valid_msb: GIG EMAC data MSByte valid bit
-- rx_good_frame: GIG EMAC good frame indicator (currently not used)
-- rx_bad_frame: GIG EMAC bad frame indicator (currently not used)
--
-- Status and clocking:
-- reset: Reset signal
-- clk: clock of client logic, ie., 2^26 Hz
-- status: Status word
-- statistics_dat: calculated time difference between provided time stamp and
--    time stamp in frame
-- statistics_sel: denotes the queue to which the frame belongs
-- statistics_rdy: a frame has been received and a time stamp difference is
--    available
--
-- Low priority queue:
-- LP_en: enable a read from the low priority queue
-- LP_d: data word from the low priority queue
-- LP_eof: denotes the last word in a low priority frame
--
-- Queue A:
-- A_active: indicates if this queue is currently in use
-- A_flush: this will empty the queue and abort an ongoing 'A' ethernet frame
-- A_tstamp: this is the time stamp associated with the current receive window
-- A_len: this is the expeted length of the A queue payload
-- A_gofirst: indicates the start of the microcode state machine
-- A_receive: this is the receive window delimiter
-- A_en: enables reading the A queue
-- A_ d: output data word
-- A_valid: indicates if output word is valid
--
-- Queues B through D same as A.
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_std.all;

entity RX_arbiter is port
   (
     rx_clk            : in    std_logic;
     rx_data           : in    std_logic_vector(15 downto 0);
     rx_data_valid     : in    std_logic;
     rx_data_valid_msb : in    std_logic;
     rx_good_frame     : in    std_logic;
     rx_bad_frame      : in    std_logic;

     reset             : in    std_logic;
     clk               : in    std_logic;
     status            : out   std_logic_vector(15 downto 0);

     statistics_dat    : out   std_logic_vector (63 downto 0);
     statistics_sel    : out   std_logic_vector (1 downto 0);
     statistics_rdy    : out   std_logic;

     LP_en             : in    std_logic;
     LP_d              : out   std_logic_vector(15 downto 0);
     LP_eof            : out   std_logic;

     A_active          : in    std_logic;
     A_flush           : in    std_logic;
     A_tstamp          : in    std_logic_vector(63 downto 0);
     A_len             : in    std_logic_vector(15 downto 0);
     A_gofirst         : in    std_logic;
     A_receive         : in    std_logic;
     A_en              : in    std_logic;
     A_d               : out   std_logic_vector(31 downto 0);
     A_valid           : out   std_logic;

     B_active          : in    std_logic;
     B_flush           : in    std_logic;
     B_tstamp          : in    std_logic_vector(63 downto 0);
     B_len             : in    std_logic_vector(15 downto 0);
     B_gofirst         : in    std_logic;
     B_receive         : in    std_logic;
     B_en              : in    std_logic;
     B_d               : out   std_logic_vector(31 downto 0);
     B_valid           : out   std_logic;

     C_active          : in    std_logic;
     C_flush           : in    std_logic;
     C_tstamp          : in    std_logic_vector(63 downto 0);
     C_len             : in    std_logic_vector(15 downto 0);
     C_gofirst         : in    std_logic;
     C_receive         : in    std_logic;
     C_en              : in    std_logic;
     C_d               : out   std_logic_vector(31 downto 0);
     C_valid           : out   std_logic;

     D_active          : in    std_logic;
     D_flush           : in    std_logic;
     D_tstamp          : in    std_logic_vector(63 downto 0);
     D_len             : in    std_logic_vector(15 downto 0);
     D_gofirst         : in    std_logic;
     D_receive         : in    std_logic;
     D_en              : in    std_logic;
     D_d               : out   std_logic_vector(31 downto 0);
     D_valid           : out   std_logic
   );
end RX_arbiter;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of RX_arbiter is

-- Type Declaration
   type rx_arbiter_state is (IDLE, EADDR, FRAMETYPE, PROTOCOLTYPE,
                             PROTOCOLVERSION, DATALENGTH, TSTAMP1,
                             TSTAMP2, TSTAMP3, TSTAMP4, PAYLOAD,
                             LOWPRIORITY, LOWPRIORITY_FINI);

-- Component Declarations
   component fifo_1024_18
      port (
         din:             IN std_logic_VECTOR(17 downto 0);
         rd_clk:          IN std_logic;
         rd_en:           IN std_logic;
         rst:             IN std_logic;
         wr_clk:          IN std_logic;
         wr_en:           IN std_logic;
         dout:            OUT std_logic_VECTOR(17 downto 0);
         empty:           OUT std_logic;
         full:            OUT std_logic;
         overflow:        OUT std_logic;
         underflow:       OUT std_logic
      );
   end component;
   component fifo_2048_18
      port (
         din:             IN std_logic_VECTOR(17 downto 0);
         rd_clk:          IN std_logic;
         rd_en:           IN std_logic;
         rst:             IN std_logic;
         wr_clk:          IN std_logic;
         wr_en:           IN std_logic;
         dout:            OUT std_logic_VECTOR(17 downto 0);
         empty:           OUT std_logic;
         full:            OUT std_logic;
         overflow:        OUT std_logic;
         underflow:       OUT std_logic
      );
   end component;

   component RX_queue
      port (
        reset             : in    std_logic;
        beg_out           : in    std_logic;
        dout              : out   std_logic_vector (31 downto 0);
        rd_clk            : in    std_logic;
        rd_en             : in    std_logic;
        rd_valid          : out   std_logic;

        beg_in            : in    std_logic;
        end_in            : in    std_logic;
        din               : in    std_logic_vector (15 downto 0);
        wr_clk            : in    std_logic;
        wr_en             : in    std_logic;
        overflow          : out   std_logic
      );
   end component RX_queue;

-- Signal Declarations
   signal state:          rx_arbiter_state;
   signal rx_data_valid_old: std_logic;
   signal addr_cntr:      unsigned (2 downto 0);
   signal len_cntr:       unsigned (12 downto 0);
   signal len_cntr_tc:    std_logic;
   signal lp_fifo_d:      std_logic_vector (17 downto 0);
   signal lp_fifo_reset:  std_logic;
   signal lp_fifo_ren:    std_logic;
   signal lp_fifo_empty:  std_logic;
   signal lp_mark_eof:    std_logic;
   signal lp_fini_cnt:    unsigned (2 downto 0);

   signal lp_out:         std_logic_vector (17 downto 0);
   signal lp_overflow:    std_logic;
   signal lp_underflow:   std_logic;
   signal lp_fifo1_in:    std_logic_vector (17 downto 0);
   signal lp_fifo2_in:    std_logic_vector (17 downto 0);

   signal rx_stop:        std_logic;
   signal queue_sel:      std_logic_vector (1 downto 0);
   signal data_len:       std_logic_vector (15 downto 0);
   signal tstamp:         std_logic_vector (63 downto 0);
   signal a_beg:          std_logic;
   signal b_beg:          std_logic;
   signal c_beg:          std_logic;
   signal d_beg:          std_logic;
   signal fifo_wr:        std_logic;
   signal a_fifo_wr:      std_logic;
   signal b_fifo_wr:      std_logic;
   signal c_fifo_wr:      std_logic;
   signal d_fifo_wr:      std_logic;
   signal flush:          std_logic;
   signal beg:            std_logic;
   signal qactive:        std_logic;
   signal a_reset:        std_logic;
   signal b_reset:        std_logic;
   signal c_reset:        std_logic;
   signal d_reset:        std_logic;
   signal ftype_err:      std_logic;
   signal len_err:        std_logic;
   signal flen_err:       std_logic;
   signal out_of_sync:    std_logic;
   signal a_out_of_sync:  std_logic;
   signal b_out_of_sync:  std_logic;
   signal c_out_of_sync:  std_logic;
   signal d_out_of_sync:  std_logic;
   signal status_1:       std_logic_vector(15 downto 0);
   signal status_2:       std_logic_vector(15 downto 0);
   signal status_3:       std_logic_vector(15 downto 0);

   signal a_overflow:     std_logic;
   signal old_a_overflow: std_logic;
   signal b_overflow:     std_logic;
   signal old_b_overflow: std_logic;
   signal c_overflow:     std_logic;
   signal old_c_overflow: std_logic;
   signal d_overflow:     std_logic;
   signal old_d_overflow: std_logic;

   signal ftime:          unsigned (63 downto 0);
   signal timediff:       unsigned (63 downto 0);
   signal longfifo_wr :   std_logic_vector (3 downto 0);
   signal stat_trig :     std_logic_vector (3 downto 0);

begin

   -- low priority FIFO for intermediate storage
   lp_fifo1_in <= B"00" & rx_data;
   LP_FIFO_inter: fifo_1024_18 port map (
      rst => lp_fifo_reset,
      din => lp_fifo1_in,
      wr_clk => rx_clk,
      wr_en => rx_data_valid,
      rd_clk => rx_clk,
      rd_en => lp_fifo_ren,
      dout => lp_fifo_d,
      empty => lp_fifo_empty,
      full => open,
      overflow => open,
      underflow => open
   );
   -- main low priority FIFO
   lp_fifo2_in <= B"0" & lp_mark_eof & lp_fifo_d (15 downto 0);
   LP_FIFO: fifo_2048_18 port map (
      rst => reset,
      din => lp_fifo2_in,
      wr_clk => rx_clk,
      wr_en => lp_fifo_ren,
      rd_clk => clk,
      rd_en => LP_en,
      dout => LP_out,
      empty => open,
      full => open,
      overflow => lp_overflow,
      underflow => lp_underflow
   );
   LP_d <= LP_out (15 downto 0);
   LP_eof <= LP_out (16);

   -- mux data length
   process (queue_sel, A_len, B_len, C_len, D_len) is
   begin
      case queue_sel is
         when B"00" =>
            data_len <= A_len;
         when B"01" =>
            data_len <= B_len;
         when B"10" =>
            data_len <= C_len;
         when others =>
            data_len <= D_len;
      end case;
   end process;

   -- mux time stamp
   process (queue_sel, A_tstamp, B_tstamp, C_tstamp, D_tstamp) is
   begin
      case queue_sel is
         when B"00" =>
            tstamp <= A_tstamp;
         when B"01" =>
            tstamp <= B_tstamp;
         when B"10" =>
            tstamp <= C_tstamp;
         when others =>
            tstamp <= D_tstamp;
      end case;
   end process;

   rx_data_valid_old <= rx_data_valid when rising_edge (clk);

   -- receive arbiter state machine
   RX_arbiter_FSM: process (rx_clk, reset, state, rx_data_valid_old,
                         rx_data_valid, rx_data, tstamp, rx_stop,
                         addr_cntr, lp_fini_cnt, data_len, len_cntr_tc,
                         len_cntr) is
   begin
      if reset = '1' then
         if rx_data_valid = '0' then
            state <= IDLE;
         end if;
         lp_fifo_reset <= '1';
         lp_fifo_ren <= '0';
         lp_fini_cnt <= B"000";
         ftype_err <= '0';
         len_err <= '0';
         out_of_sync <= '0';
         fifo_wr <= '0';
         beg <= '0';
      elsif rising_edge (rx_clk) then
         case state is
            -- idle state: check rx_data_valid
            when IDLE =>
               lp_fifo_reset <= '0';
               ftype_err <= '0';
               len_err <= '0';
               out_of_sync <= '0';
               fifo_wr <= '0';
               beg <= '0';
               if (rx_data_valid = '1') and (rx_data_valid_old = '0') then
                  state <= EADDR;
                  addr_cntr <= B"010";
               end if;
            -- receive ethernet address
            when EADDR =>
               if rx_data_valid = '0' then
                  state <= IDLE;
                  lp_fifo_reset <= '1';
               elsif addr_cntr = B"110" then
                  state <= FRAMETYPE;
               end if;
               addr_cntr <= addr_cntr + 1;
            -- receive frame type
            when FRAMETYPE =>
               if (rx_data_valid = '0') or (rx_data /= X"88B5") then
                  state <= IDLE;
                  ftype_err <= '1';
               else
                  state <= PROTOCOLTYPE;
               end if;
            -- receive protocol subtype
            when PROTOCOLTYPE =>
               if (rx_data_valid = '0') then
                  state <= IDLE;
                  lp_fifo_reset <= '1';
               elsif (rx_data(15) = '1') then
                  queue_sel <= B"00";
                  state <= PROTOCOLVERSION;
               elsif (rx_data(14) = '1') then
                  queue_sel <= B"01";
                  state <= PROTOCOLVERSION;
               elsif (rx_data(13) = '1') then
                  queue_sel <= B"10";
                  state <= PROTOCOLVERSION;
               elsif (rx_data(12) = '1') then
                  queue_sel <= B"11";
                  state <= PROTOCOLVERSION;
               else
                  state <= LOWPRIORITY;
                  lp_fifo_ren <= '1';
               end if;
            -- receive protocol version
            when PROTOCOLVERSION =>
               lp_fifo_reset <= '1';
               if (rx_stop = '1') then
                  state <= IDLE;
               elsif (rx_data /= X"0000") then
                  state <= IDLE;
                  ftype_err <= '1';
               else
                  state <= DATALENGTH;
               end if;
            -- receive data length
            when DATALENGTH =>
               if (rx_stop = '1') then
                  state <= IDLE;
               elsif (rx_data (13 downto 1) = B"0000000000000") then
                  state <= IDLE;
                  len_err <= '1';
               else
                  state <= TSTAMP1;
                  len_cntr <= unsigned (rx_data (13 downto 1));
               end if;
            -- receive time stamp, 1st word
            when TSTAMP1 =>
               ftime (63 downto 48) <= unsigned (rx_data);
               if (rx_stop = '1') then
                  state <= IDLE;
               elsif (rx_data /= tstamp (63 downto 48)) then
                  state <= IDLE;
                  out_of_sync <= '1';
               elsif len_cntr_tc = '1' then
                  state <= IDLE;
                  len_err <= '1';
               else
                  state <= TSTAMP2;
               end if;
               len_cntr <= len_cntr - 1;
            -- receive time stamp, 2nd word
            when TSTAMP2 =>
               ftime (47 downto 32) <= unsigned (rx_data);
               if (rx_stop = '1') then
                  state <= IDLE;
               elsif (rx_data /= tstamp (47 downto 32)) then
                  state <= IDLE;
                  out_of_sync <= '1';
               elsif len_cntr_tc = '1' then
                  state <= IDLE;
                  len_err <= '1';
               else
                  state <= TSTAMP3;
               end if;
               len_cntr <= len_cntr - 1;
            -- receive time stamp, 3rd word
            when TSTAMP3 =>
               ftime (31 downto 16) <= unsigned (rx_data);
               if (rx_stop = '1') then
                  state <= IDLE;
               elsif (rx_data /= tstamp (31 downto 16)) then
                  state <= IDLE;
                  out_of_sync <= '1';
               elsif len_cntr_tc = '1' then
                  state <= IDLE;
                  len_err <= '1';
               else
                  state <= TSTAMP4;
                  beg <= '1';
               end if;
               len_cntr <= len_cntr - 1;
            -- receive time stamp, 4th word
            when TSTAMP4 =>
               ftime (15 downto 0) <= unsigned (rx_data);
               -- if there is an error here beg will be applied twice
               -- and toggle the buffer forth and back!
               if (rx_stop = '1') then
                  state <= IDLE;
               elsif (rx_data /= tstamp (15 downto 0)) then
                  state <= IDLE;
                  out_of_sync <= '1';
               elsif len_cntr_tc = '1' then
                  state <= IDLE;
                  len_err <= '1';
               else
                  fifo_wr <= '1';
                  beg <= '0';  -- toggle only once when ok
                  state <= PAYLOAD;
               end if;
               len_cntr <= len_cntr - 1;
            -- transfer payload
            when PAYLOAD =>
               if (rx_stop = '1') then
                  state <= IDLE;
                  len_err <= '1';
                  fifo_wr <= '0';
               elsif (len_cntr_tc = '1') then
                  state <= IDLE;
                  fifo_wr <= '0';
               end if;
               len_cntr <= len_cntr - 1;
            -- transfer zeros due to payload error
            -- transfer low priority ethernet frame into queue
            when LOWPRIORITY =>
               if (rx_data_valid = '0') then
                  state <= LOWPRIORITY_FINI;
                  lp_fini_cnt <= B"001";
               end if;
            -- finish low priority transfer
            when LOWPRIORITY_FINI =>
               if lp_fini_cnt = B"111" then
                  state <= IDLE;
                  lp_fifo_ren <= '0';
               end if;
               lp_fini_cnt <= lp_fini_cnt + 1;
         end case;
      end if;
   end process RX_arbiter_FSM;
   -- stop receiving data frames
   rx_stop <= '1' when (rx_data_valid = '0') or (flush = '1') or
                       (qactive = '0') else '0';
   -- end of frame mark
   lp_mark_eof <= '1' when lp_fini_cnt = B"111" else '0';
   -- end of payload mark
   len_cntr_tc <= '1' when len_cntr = B"0000000000001" else '0';

   -- A queue FIFO
   a_reset <= reset or a_flush;
   A_queue: RX_queue port map (
      reset => a_reset,
      beg_out => A_gofirst,
      dout => A_d,
      rd_clk => clk,
      rd_en => A_en,
      rd_valid => A_valid,
      beg_in => a_beg,
      end_in => A_receive,
      din => rx_data,
      wr_clk => rx_clk,
      wr_en => a_fifo_wr,
      overflow => a_overflow
     );
   -- B queue FIFO
   b_reset <= reset or b_flush;
   B_queue: RX_queue port map (
      reset => b_reset,
      beg_out => B_gofirst,
      dout => B_d,
      rd_clk => clk,
      rd_en => B_en,
      rd_valid => B_valid,
      beg_in => b_beg,
      end_in => B_receive,
      din => rx_data,
      wr_clk => rx_clk,
      wr_en => b_fifo_wr,
      overflow => b_overflow
     );
   -- C queue FIFO
   c_reset <= reset or c_flush;
   C_queue: RX_queue port map (
      reset => c_reset,
      beg_out => C_gofirst,
      dout => C_d,
      rd_clk => clk,
      rd_en => C_en,
      rd_valid => C_valid,
      beg_in => c_beg,
      end_in => C_receive,
      din => rx_data,
      wr_clk => rx_clk,
      wr_en => c_fifo_wr,
      overflow => c_overflow
     );
   -- D queue FIFO
   d_reset <= reset or d_flush;
   D_queue: RX_queue port map (
      reset => d_reset,
      beg_out => D_gofirst,
      dout => D_d,
      rd_clk => clk,
      rd_en => D_en,
      rd_valid => D_valid,
      beg_in => d_beg,
      end_in => D_receive,
      din => rx_data,
      wr_clk => rx_clk,
      wr_en => d_fifo_wr,
      overflow => d_overflow
     );

   -- demux high priority FIFO write enable
   process (queue_sel, fifo_wr) is
   begin
      case queue_sel is
         when B"00" =>
            A_fifo_wr <= fifo_wr;
            B_fifo_wr <= '0';
            C_fifo_wr <= '0';
            D_fifo_wr <= '0';
         when B"01" =>
            A_fifo_wr <= '0';
            B_fifo_wr <= fifo_wr;
            C_fifo_wr <= '0';
            D_fifo_wr <= '0';
         when B"10" =>
            A_fifo_wr <= '0';
            B_fifo_wr <= '0';
            C_fifo_wr <= fifo_wr;
            D_fifo_wr <= '0';
         when others =>
            A_fifo_wr <= '0';
            B_fifo_wr <= '0';
            C_fifo_wr <= '0';
            D_fifo_wr <= fifo_wr;
      end case;
   end process;
   -- demux high priority FIFO begin
   process (queue_sel, beg) is
   begin
      case queue_sel is
         when B"00" =>
            A_beg <= beg;
            B_beg <= '0';
            C_beg <= '0';
            D_beg <= '0';
         when B"01" =>
            A_beg <= '0';
            B_beg <= beg;
            C_beg <= '0';
            D_beg <= '0';
         when B"10" =>
            A_beg <= '0';
            B_beg <= '0';
            C_beg <= beg;
            D_beg <= '0';
         when others =>
            A_beg <= '0';
            B_beg <= '0';
            C_beg <= '0';
            D_beg <= beg;
      end case;
   end process;
   -- mux flush
   process (queue_sel, A_flush, B_flush, C_flush, D_flush) is
   begin
      case queue_sel is
         when B"00" =>
            flush <= A_flush;
         when B"01" =>
            flush <= B_flush;
         when B"10" =>
            flush <= C_flush;
         when others =>
            flush <= D_flush;
      end case;
   end process;
   -- mux active
   process (queue_sel, A_active, B_active, C_active, D_active) is
   begin
      case queue_sel is
         when B"00" =>
            qactive <= A_active;
         when B"01" =>
            qactive <= B_active;
         when B"10" =>
            qactive <= C_active;
         when others =>
            qactive <= D_active;
      end case;
   end process;
   -- demux out of sync error
   process (queue_sel, out_of_sync) is
   begin
      case queue_sel is
         when B"00" =>
            A_out_of_sync <= out_of_sync;
            B_out_of_sync <= '0';
            C_out_of_sync <= '0';
            D_out_of_sync <= '0';
         when B"01" =>
            A_out_of_sync <= '0';
            B_out_of_sync <= out_of_sync;
            C_out_of_sync <= '0';
            D_out_of_sync <= '0';
         when B"10" =>
            A_out_of_sync <= '0';
            B_out_of_sync <= '0';
            C_out_of_sync <= out_of_sync;
            D_out_of_sync <= '0';
         when others =>
            A_out_of_sync <= '0';
            B_out_of_sync <= '0';
            C_out_of_sync <= '0';
            D_out_of_sync <= out_of_sync;
      end case;
   end process;
   -- check overflow
   old_a_overflow <= a_overflow when rising_edge (rx_clk);
   old_b_overflow <= b_overflow when rising_edge (rx_clk);
   old_c_overflow <= c_overflow when rising_edge (rx_clk);
   old_d_overflow <= d_overflow when rising_edge (rx_clk);
   flen_err <= len_err or (a_overflow and not old_a_overflow) or
               (b_overflow and not old_b_overflow) or
               (c_overflow and not old_c_overflow) or
               (d_overflow and not old_d_overflow);
   -- set status
   status_1 <= B"000000000" & flen_err & ftype_err &
               d_out_of_sync & c_out_of_sync &
               b_out_of_sync & a_out_of_sync & B"0";
   -- now resync to clk
   status_2 <= status_1 when rising_edge (clk);
   status_3 <= status_2 when rising_edge (clk);
   status <= (status_2(15 downto 1) and not status_3(15 downto 1)) &
             not lp_fifo_empty;

   -- gather statistics
   timediff <= ftime - unsigned (tstamp);
   process (rx_clk, fifo_wr, longfifo_wr) is
   begin
      if rising_edge (rx_clk) then
         if fifo_wr = '1' then
            longfifo_wr <= B"1111";
         else
            longfifo_wr <= longfifo_wr (2 downto 0) & '0';
         end if;
      end if;
   end process;
   -- now resync to clk
   stat_trig <= stat_trig (2 downto 0) & longfifo_wr(3) when rising_edge (clk);
   statistics_rdy <= not stat_trig (3) and stat_trig (2);
   statistics_dat <= std_logic_vector (timediff);
   statistics_sel <= queue_sel;

end Structure;
--------------------------------------------------------------------------------
