--------------------------------------------------------------------------------
-- Stand Alone Uplink: FPGA Design
-- LIGO document: D060288
-- SubModule RX_queue
-- Created   9/25/2006 5:02:25 PM
--
-- RX_queue implements a buffer for the isosynchronous receive queues.
-- It uses double buffering to switch between reading and writing. The write
-- buffer switches whenever a new valid frame has been received and this
-- buffer then becomes the read buffer at the end of the receive window.
-- The buffers are read in the same order as they are read. In this reagard
-- they behave like true FIFOs. The main difference is that the buffer
-- switching guarantees that the data stays in synchronization.
--
-- Data of a full frame must be written without a break.
--
-- Parameters:
-- reset: reset input
-- beg_out: begin of the read process; must be the start of the microcode
--    state machine
-- dout: output data dword
-- rd_clk: read clock (user logic)
-- rd_en: enable read
-- rd_valid: indicates that the read data is valid
--
-- beg_in: begin of the write process (start of data copy)
-- end_in: end of the write process (receive window)
-- din: input data word
-- wr_clk: write clock (EMAC)
-- wr_en: enable write
-- overflow: buffer has overflown
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_std.all;

entity RX_queue is port
   (
     reset             : in    std_logic;
     beg_out           : in    std_logic;  -- start microcode state machine
     dout              : out   std_logic_vector (31 downto 0);
     rd_clk            : in    std_logic;
     rd_en             : in    std_logic;
     rd_valid          : out   std_logic;

     beg_in            : in    std_logic;  -- start data copy into queue
     end_in            : in    std_logic;  -- end of receive window
     din               : in    std_logic_vector (15 downto 0);
     wr_clk            : in    std_logic;
     wr_en             : in    std_logic;
     overflow          : out   std_logic
   );
end RX_queue;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of RX_queue is

-- Type Declaration

-- Component Declarations
   component bram_2048_18_36
       port (
       clka: IN std_logic;
       dina: IN std_logic_VECTOR(17 downto 0);
       addra: IN std_logic_VECTOR(10 downto 0);
       ena: IN std_logic;
       wea: IN std_logic_VECTOR(0 downto 0);
       douta: OUT std_logic_VECTOR(17 downto 0);
       clkb: IN std_logic;
       dinb: IN std_logic_VECTOR(35 downto 0);
       addrb: IN std_logic_VECTOR(9 downto 0);
       enb: IN std_logic;
       web: IN std_logic_VECTOR(0 downto 0);
       doutb: OUT std_logic_VECTOR(35 downto 0));
   end component;

   component bram_4096_18_36
       port (
       clka: IN std_logic;
       dina: IN std_logic_VECTOR(17 downto 0);
       addra: IN std_logic_VECTOR(11 downto 0);
       wea: IN std_logic_VECTOR(0 downto 0);
       douta: OUT std_logic_VECTOR(17 downto 0);
       clkb: IN std_logic;
       dinb: IN std_logic_VECTOR(35 downto 0);
       addrb: IN std_logic_VECTOR(10 downto 0);
       web: IN std_logic_VECTOR(0 downto 0);
       doutb: OUT std_logic_VECTOR(35 downto 0));
   end component;

-- Signal Declarations
   signal reset_buf    : std_logic_vector (3 downto 0);
   signal longreset    : std_logic;
   signal buf_in       : std_logic;
   signal buf_out      : std_logic;
   signal next_buf_out : std_logic;
   signal queue_dina   : std_logic_vector (17 downto 0);
   signal queue_doutb  : std_logic_vector (35 downto 0);
   signal queue_adra   : std_logic_vector (11 downto 0);
   signal queue_adrb   : std_logic_vector (10 downto 0);
   signal queue_wea    : std_logic_vector (0 downto 0);
   signal in_addr      : unsigned (10 downto 0);
   signal in_addr_tc   : std_logic;
   signal in_addr_valid : std_logic;
   signal out_addr     : unsigned (9 downto 0);
   signal out_valid    : std_logic;
   signal invalid      : std_logic;
   signal out_addr_tc  : std_logic;
   signal full         : std_logic;
   signal empty        : std_logic;
   signal underflow    : std_logic;

begin
   -- prolong reset
   process (rd_clk, reset, reset_buf) is
   begin
      if reset = '1' then
         reset_buf <= B"1111";
      elsif rising_edge (rd_clk) then
         reset_buf <= reset_buf (2 downto 0) & '0';
      end if;
   end process;
   longreset <= reset_buf (3);

   -- determine which buffer to use for write
   process (wr_clk, beg_in, end_in, beg_in, buf_out) is
   begin
      if rising_edge (wr_clk) then
         if beg_in = '1' then
            buf_in <= not buf_in;
         end if;
      end if;
   end process;
   -- determine which buffer to use for read
   process (rd_clk, beg_out, end_in, buf_in, next_buf_out) is
   begin
      if rising_edge (rd_clk) then
         if end_in = '1' then
            next_buf_out <= buf_in;
         end if;
         if beg_out = '1' then
            buf_out <= next_buf_out;
         end if;
      end if;
   end process;

   -- input address
   process (longreset, wr_clk, beg_in, wr_en, in_addr, in_addr_tc, full) is
   begin
      if longreset = '1' then
         in_addr <= (others => '0');
         full <= '0';
         overflow <= '0';
      elsif rising_edge (wr_clk) then
         overflow <= '0';
         if beg_in = '1' then
            in_addr <= (others => '0');
            full <= '0';
         elsif wr_en = '1' then
            if in_addr_tc = '1' then
               full <= '1';
               overflow <= full;
            else
               in_addr <= in_addr + 1;
            end if;
         end if;
      end if;
   end process;
   in_addr_tc <= '1' when in_addr = B"111111111111" else '0';
   in_addr_valid <= wr_en;

   -- output address
   process (longreset, rd_clk, beg_out, rd_en, out_addr, out_addr_tc, empty) is
   begin
      if longreset = '1' then
         out_addr <= (others => '0');
         empty <= '0';
         underflow <= '0';
      elsif rising_edge (rd_clk) then
         underflow <= '0';
         if beg_out = '1' then
            out_addr <= (others => '0');
            empty <= '0';
         elsif rd_en = '1' then
            if out_addr_tc = '1' then
               empty <= '1';
               underflow <= empty;
            else
               out_addr <= out_addr + 1;
            end if;
         end if;
      end if;
   end process;
   out_addr_tc <= '1' when out_addr = B"11111111111" else '0';

   -- dual buffer dual port ram block
   queue_dina <= B"0" & in_addr_valid & din;
   queue_adra <= buf_in & std_logic_vector (in_addr);
   queue_adrb <= buf_out & std_logic_vector (out_addr);
   queue_wea(0) <= not full;
   queue: bram_4096_18_36 port map
   (
      clka => wr_clk,
      dina => queue_dina,
      addra => queue_adra,
      wea => queue_wea,
      douta => open,
      clkb => rd_clk,
      dinb => X"000000000",
      addrb => queue_adrb,
      web => B"0",
      doutb => queue_doutb
   );
   dout <= queue_doutb (33 downto 18) & queue_doutb (15 downto 0);
   out_valid <= queue_doutb (16) and queue_doutb (32);

   -- latch valid
   process (longreset, rd_clk, beg_out, out_valid) is
   begin
      if (longreset = '1') then
         invalid <= '1';
      elsif rising_edge (rd_clk) then
         if (beg_out = '1') then
            invalid <= '0';
         elsif out_valid = '0' then
            invalid <= '1';
         end if;
      end if;
   end process;
   rd_valid <= out_valid and not invalid and not underflow;

end architecture Structure;
--------------------------------------------------------------------------------









