--------------------------------------------------------------------------------
-- Stand Alone Uplink: FPGA Design
-- LIGO document: D060288
-- SubModule TX_datasel
-- Created   9/15/2006 7:37:05 PM
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;

entity TX_datasel is port
   (
     CLK            : in    std_logic;
     ADC            : in    std_logic_vector(31 downto 0);
     MP             : in    std_logic_vector(31 downto 0);
     GPSsec         : in    std_logic_vector(31 downto 0);
     GPSfrac        : in    std_logic_vector(25 downto 0);
     ADR            : in    std_logic_vector(11 downto 8);
     SEL            : in    std_logic_vector(1 downto 0);
     A_time         : in    std_logic;
     B_time         : in    std_logic;
     C_time         : in    std_logic;
     D_time         : in    std_logic;
     O              : out   std_logic_vector(31 downto 0)
   );
end TX_datasel;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TX_datasel is

-- Component Declarations

-- Signal Declarations
   signal a         : std_logic_vector (2 downto 0);
   signal b         : std_logic;
   signal c         : std_logic;
   signal muxout    : std_logic_vector (31 downto 0);
   signal reg       : std_logic_vector (31 downto 0);
   signal w         : std_logic_vector (31 downto 0);

   signal A_GPSs    : std_logic_vector (31 downto 0);
   signal A_GPSf    : std_logic_vector (31 downto 0);
   signal B_GPSs    : std_logic_vector (31 downto 0);
   signal B_GPSf    : std_logic_vector (31 downto 0);
   signal C_GPSs    : std_logic_vector (31 downto 0);
   signal C_GPSf    : std_logic_vector (31 downto 0);
   signal D_GPSs    : std_logic_vector (31 downto 0);
   signal D_GPSf    : std_logic_vector (31 downto 0);
   signal GPSs      : std_logic_vector (31 downto 0);
   signal GPSf      : std_logic_vector (31 downto 0);

begin
   -- GPS time stamp
   process (CLK, A_GPSs, A_GPSf, A_time) is
   begin
      if rising_edge (CLK) then
         if A_time = '1' then
            A_GPSs <= GPSsec;
            A_GPSf <= GPSfrac(25 downto 0) & B"000000";
         end if;
      end if;
   end process;
   process (CLK, B_GPSs, B_GPSf, B_time) is
   begin
      if rising_edge (CLK) then
         if B_time = '1' then
            B_GPSs <= GPSsec;
            B_GPSf <= GPSfrac & B"000000";
         end if;
      end if;
   end process;
   process (CLK, C_GPSs, C_GPSf, C_time) is
   begin
      if rising_edge (CLK) then
         if C_time = '1' then
            C_GPSs <= GPSsec;
            C_GPSf <= GPSfrac & B"000000";
         end if;
      end if;
   end process;
   process (CLK, D_GPSs, D_GPSf, D_time) is
   begin
      if rising_edge (CLK) then
         if D_time = '1' then
            D_GPSs <= GPSsec;
            D_GPSf <= GPSfrac & B"000000";
         end if;
      end if;
   end process;

   -- GPS time stamp mux
   process (SEL, A_GPSs, B_GPSs, C_GPSs, D_GPSs,
            A_GPSf, B_GPSf, C_GPSf, D_GPSf) is
   begin
      case SEL is
         when B"00" =>
            GPSs <= A_GPSs;
            GPSf <= A_GPSf;
         when B"01" =>
            GPSs <= B_GPSs;
            GPSf <= B_GPSf;
         when B"10" =>
            GPSs <= C_GPSs;
            GPSf <= C_GPSf;
         when others =>
            GPSs <= D_GPSs;
            GPSf <= D_GPSf;
      end case;
   end process;

   -- delay address by one clock cycle to account for pipeline delay
   a <= ADR(10 downto 8) when rising_edge (CLK);

   -- mux data memory pool and GPS sec
   process (a, MP, GPSs, GPSf) is
   begin
      case a is
         when B"000" | B"001" =>
            muxout <= MP;
         when B"010" =>
            muxout <= GPSs;
         when B"011" =>
            muxout <= GPSf;
         when B"111" =>
            muxout <= (others => '0');
         when others =>
            muxout <= (others => '0');
      end case;
   end process;
   -- load register with mux output
   reg <= muxout when rising_edge (CLK);

   -- delay ADR(11) by 2 clock cycle to account for pipeline delay
   b <= ADR(11) when rising_edge (CLK);
   c <= b when rising_edge (CLK);

   -- mux memory pool against ADC
   process (c, reg, ADC) is
   begin
      case c is
         when '0' =>
            w <= ADC(31 downto 0);
         when others =>
            w <= reg(31 downto 0);
      end case;
   end process;

   -- set output
   O <= w;

end Structure;
--------------------------------------------------------------------------------
