--------------------------------------------------------------------------------
-- Stand Alone Uplink: FPGA Design
-- LIGO document: D060288
-- SubModule TX_pio_stat_WB
-- Created   9/25/2006 11:44:00 AM
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_std.all;

entity TX_pio_stat_WB is port
   (
     CLK                : in    std_logic;

     STB_I              : in    std_logic;
     CYC_I              : in    std_logic;
     ACK_O              : out   std_logic;
     ADR_I              : in    std_logic_vector(4 downto 0);
     DAT_I              : in    std_logic_vector(31 downto 0);
     DAT_O              : out   std_logic_vector(31 downto 0);
     WE_I               : in    std_logic;
     CLK_I              : in    std_logic;
     RST_I              : in    std_logic;

     FIFO_EN            : out   std_logic;
     FIFO_O             : out   std_logic_vector(15 downto 0);
     FIFO_EOF           : out   std_logic;

     LPSTART            : out   std_logic_vector(31 downto 0);
     LPSTOP             : out   std_logic_vector(31 downto 0);
     LPMASK             : out   std_logic_vector(31 downto 0);
     ATRIG              : out   std_logic_vector(31 downto 0);
     BTRIG              : out   std_logic_vector(31 downto 0);
     CTRIG              : out   std_logic_vector(31 downto 0);
     DTRIG              : out   std_logic_vector(31 downto 0);
     AMASK              : out   std_logic_vector(31 downto 0);
     BMASK              : out   std_logic_vector(31 downto 0);
     CMASK              : out   std_logic_vector(31 downto 0);
     DMASK              : out   std_logic_vector(31 downto 0);
     AENDM              : out   std_logic_vector(31 downto 0);
     DENDM              : out   std_logic_vector(31 downto 0);
     BENDM              : out   std_logic_vector(31 downto 0);
     CENDM              : out   std_logic_vector(31 downto 0);

     statArbiter        : in    std_logic_vector(15 downto 0);
     statFSM            : in    std_logic_vector(15 downto 0)
   );
end TX_pio_stat_WB;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TX_pio_stat_WB is

-- Component Declarations

-- Signal Declarations
   signal atrig_i : std_logic_vector(31 downto 0) := (others => '0');
   signal btrig_i : std_logic_vector(31 downto 0) := (others => '0');
   signal ctrig_i : std_logic_vector(31 downto 0) := (others => '0');
   signal dtrig_i : std_logic_vector(31 downto 0) := (others => '0');
   signal amask_i : std_logic_vector(31 downto 0) := (others => '0');
   signal bmask_i : std_logic_vector(31 downto 0) := (others => '0');
   signal cmask_i : std_logic_vector(31 downto 0) := (others => '0');
   signal dmask_i : std_logic_vector(31 downto 0) := (others => '0');
   signal aendm_i : std_logic_vector(31 downto 0) := (others => '0');
   signal bendm_i : std_logic_vector(31 downto 0) := (others => '0');
   signal cendm_i : std_logic_vector(31 downto 0) := (others => '0');
   signal dendm_i : std_logic_vector(31 downto 0) := (others => '0');
   signal lpstart_i : std_logic_vector(31 downto 0) := (others => '0');
   signal lpstop_i : std_logic_vector(31 downto 0) := (others => '0');
   signal lpmask_i : std_logic_vector(31 downto 0) := (others => '0');

   signal dat_o_i : std_logic_vector(31 downto 0);

   signal cnt_afifo_err : unsigned (31 downto 0);
   signal cnt_bfifo_err : unsigned (31 downto 0);
   signal cnt_cfifo_err : unsigned (31 downto 0);
   signal cnt_dfifo_err : unsigned (31 downto 0);
   signal cnt_lpfifo_err : unsigned (31 downto 0);
   signal cnt_arbiter_err : unsigned (31 downto 0);
   signal cnt_decod_err : unsigned (31 downto 0);

begin
   -- acknowledge
   ACK_O <= STB_I and CYC_I;

   -- Read data
   readdata: process (ADR_I, STB_I, CYC_I, atrig_i, btrig_i, ctrig_i, dtrig_i,
                      amask_i, bmask_i, cmask_i, dmask_i,
                      aendm_i, bendm_i, cendm_i, dendm_i,
                      lpstart_i, lpstop_i, lpmask_i,
                      cnt_afifo_err, cnt_bfifo_err, cnt_cfifo_err,
                      cnt_dfifo_err, cnt_lpfifo_err, cnt_arbiter_err,
                      cnt_decod_err, statArbiter, statFSM) is
   begin
      if (STB_I = '1') and (CYC_I = '1') then
         case ADR_I is
            when B"00000" =>
               dat_o_i <= atrig_i;
            when B"00001" =>
               dat_o_i <= btrig_i;
            when B"00010" =>
               dat_o_i <= ctrig_i;
            when B"00011" =>
               dat_o_i <= dtrig_i;
            when B"00100" =>
               dat_o_i <= amask_i;
            when B"00101" =>
               dat_o_i <= bmask_i;
            when B"00110" =>
               dat_o_i <= cmask_i;
            when B"00111" =>
               dat_o_i <= dmask_i;
            when B"01000" =>
               dat_o_i <= aendm_i;
            when B"01001" =>
               dat_o_i <= bendm_i;
            when B"01010" =>
               dat_o_i <= cendm_i;
            when B"01011" =>
               dat_o_i <= dendm_i;
            when B"01100" =>
               dat_o_i <= lpstart_i;
            when B"01101" =>
               dat_o_i <= lpstop_i;
            when B"01110" =>
               dat_o_i <= lpmask_i;
            when B"01111" =>
               dat_o_i <= statFSM & statArbiter;
            when B"10000" =>
               dat_o_i <= std_logic_vector (cnt_afifo_err);
            when B"10001" =>
               dat_o_i <= std_logic_vector (cnt_bfifo_err);
            when B"10010" =>
               dat_o_i <= std_logic_vector (cnt_cfifo_err);
            when B"10011" =>
               dat_o_i <= std_logic_vector (cnt_dfifo_err);
            when B"10100" =>
               dat_o_i <= std_logic_vector (cnt_lpfifo_err);
            when B"10101" =>
               dat_o_i <= std_logic_vector (cnt_arbiter_err);
            when B"10110" =>
               dat_o_i <= std_logic_vector (cnt_decod_err);
            when others =>
               dat_o_i <= (others => '0');
         end case;
      else
         dat_o_i <= (others => '0');
      end if;
   end process readdata;
   DAT_O <= dat_o_i;

   -- Write cycle
   writedata: process (CLK_I, RST_I, ADR_I, DAT_I, WE_I) is
   begin
      if rising_edge (CLK_I) then
         if RST_I = '1' then
            atrig_i <= (others => '0');
            btrig_i <= (others => '0');
            ctrig_i <= (others => '0');
            dtrig_i <= (others => '0');
            amask_i <= (others => '0');
            bmask_i <= (others => '0');
            cmask_i <= (others => '0');
            dmask_i <= (others => '0');
            aendm_i <= (others => '0');
            bendm_i <= (others => '0');
            cendm_i <= (others => '0');
            dendm_i <= (others => '0');
            lpstart_i <= (others => '0');
            lpstop_i <= (others => '0');
            lpmask_i <= (others => '0');
         elsif (STB_I = '1') and (CYC_I = '1') and (WE_I = '1') then
            case ADR_I is
               when B"00000" =>
                  atrig_i <= DAT_I;
               when B"00001" =>
                  btrig_i <= DAT_I;
               when B"00010" =>
                  ctrig_i <= DAT_I;
               when B"00011" =>
                  dtrig_i <= DAT_I;
               when B"00100" =>
                  amask_i <= DAT_I;
               when B"00101" =>
                  bmask_i <= DAT_I;
               when B"00110" =>
                  cmask_i <= DAT_I;
               when B"00111" =>
                  dmask_i <= DAT_I;
               when B"01000" =>
                  aendm_i <= DAT_I;
               when B"01001" =>
                  bendm_i <= DAT_I;
               when B"01010" =>
                  cendm_i <= DAT_I;
               when B"01011" =>
                  dendm_i <= DAT_I;
               when B"01100" =>
                  lpstart_i <= DAT_I;
               when B"01101" =>
                  lpstop_i <= DAT_I;
               when B"01110" =>
                  lpmask_i <= DAT_I;
               when B"01111" =>
                  -- FIFO write for this address
               when others =>
                  -- nothing
            end case;
         end if;
      end if;
   end process writedata;

   -- FIFO write cycle
   FIFO_O <= DAT_I (15 downto 0);
   FIFO_EOF <= DAT_I (16);
   FIFO_EN <= '1' when (WE_I = '1') and (ADR_I = B"01111") else '0';

   -- export register signals
   LPSTART <= lpstart_i;
   LPSTOP <= lpstop_i;
   LPMASK <= lpmask_i;
   ATRIG <= atrig_i;
   BTRIG <= btrig_i;
   CTRIG <= ctrig_i;
   DTRIG <= dtrig_i;
   AMASK <= amask_i;
   BMASK <= bmask_i;
   CMASK <= cmask_i;
   DMASK <= dmask_i;
   AENDM <= aendm_i;
   DENDM <= bendm_i;
   BENDM <= cendm_i;
   CENDM <= dendm_i;

   -- Statistics counter for arbiter and microcode state machine
   stat_arbiter: process (CLK, statArbiter, statFSM) is
   begin
      if rising_edge (CLK) then
         if RST_I = '1' then
            cnt_afifo_err <= (others => '0');
            cnt_bfifo_err <= (others => '0');
            cnt_cfifo_err <= (others => '0');
            cnt_dfifo_err <= (others => '0');
            cnt_lpfifo_err <= (others => '0');
            cnt_arbiter_err <= (others => '0');
            cnt_decod_err <= (others => '0');
         else
            if (statArbiter(0) = '1') or
               (statArbiter(1) = '1') or
               (statArbiter(2) = '1') then
               cnt_afifo_err <= cnt_afifo_err + 1;
            end if;
            if (statArbiter(3) = '1') or
               (statArbiter(4) = '1') or
               (statArbiter(5) = '1') then
               cnt_bfifo_err <= cnt_bfifo_err + 1;
            end if;
            if (statArbiter(6) = '1') or
               (statArbiter(7) = '1') or
               (statArbiter(8) = '1') then
               cnt_cfifo_err <= cnt_cfifo_err + 1;
            end if;
            if (statArbiter(9) = '1') or
               (statArbiter(10) = '1') or
               (statArbiter(11) = '1') then
               cnt_dfifo_err <= cnt_dfifo_err + 1;
            end if;
            if (statArbiter(12) = '1') or
               (statArbiter(13) = '1') or
               (statArbiter(14) = '1') then
               cnt_lpfifo_err <= cnt_lpfifo_err + 1;
            end if;
            if (statArbiter(15) = '1') then
               cnt_arbiter_err <= cnt_arbiter_err + 1;
            end if;
            if (statFSM(0) = '1') then
               cnt_decod_err <= cnt_decod_err + 1;
            end if;
         end if;
      end if;
   end process stat_arbiter;

end Structure;
--------------------------------------------------------------------------------
