--------------------------------------------------------------------------------
-- Stand Alone Uplink: FPGA Design
-- LIGO document: D060288
-- SubModule TrigQueue
-- Created   9/14/2006 6:56:57 PM
--
-- TrigQueue generates triggers which are synchronized with a GPS clock
-- Triggers will repeat themselves every 1 second and trigger rates have
-- to be of the form 2^-n Hz. A series of triggers can the be bundled
-- together with special markings for the first and last. The number of
-- bundled triggers has to be of the form 2^i.
--
-- The trigger repetition rate is determined by MASK.
--    GO <= (CNT and MASK) = ((TRIG+1) and MASK)
-- Triggers are bundled by ENDM so that
--    LAST <= (CNT and ENDM) = ((TRIG+1) and ENDM)
-- If ENDM = MASK, the bundle consists of a single trigger. The trigger
-- following LAST is always marked as FIRST. The value TRIG is used to offset
-- the trigger time within trigger period. It marks the LAST trigger in the
-- bundle!
--
-- Parameters:
-- CLK: clock
-- ENABLE: triggering is enabled when '1', otherwise disabled
-- CNT: 26 bits of GPS fractional seconds (units are 2^-26s)
-- TRIG: trigger time of last trigger in bundle
-- MASK: Determines the trigger repetition rate
-- ENDM: Determines the bundling of trigger
-- GO: Trigger (applied for one cycle)
-- FIRST: Marks the first trigger in a bundle (applied for one cycle)
-- LAST: Marks the last trigger in a bundle (applied for one cycle)
-- FLUSH: Flush will be applied to indicate that the triggering has been
--        disabled. It should be used to flush buffers and reset the state
--        machines which process the triggered data (applied for one cycle)
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;

entity TrigQueue is port
   (
     CLK         : in    std_logic;
     GO          : out   std_logic;
     FIRST       : out   std_logic;
     LAST        : out   std_logic;
     FLUSH       : out   std_logic;
     ENABLE      : in    std_logic;
     CNT         : in    std_logic_vector(25 downto 0);
     TRIG        : in    std_logic_vector(25 downto 0);
     MASK        : in    std_logic_vector(25 downto 0);
     ENDM        : in    std_logic_vector(25 downto 0)
   );
end TrigQueue;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TrigQueue is

-- Signal Declarations
   signal c    : std_logic_vector(25 downto 0);
   signal t    : std_logic_vector(25 downto 0);
   signal ec   : std_logic_vector(25 downto 0);
   signal et   : std_logic_vector(25 downto 0);
   signal cond : std_logic := '0';
   signal begcond : std_logic := '0';
   signal endcond : std_logic := '0';
   signal en : std_logic;
   signal olden : std_logic_vector (1 downto 0);
   signal newen : std_logic := '0';

begin
   c <= MASK and CNT ;
   t <= MASK and TRIG;
   ec <= ENDM and CNT;
   et <= ENDM and TRIG;

   en <= ENABLE;
   process (CLK, cond, t, c, ec, et, endcond, en) is
   begin
      if en = '0' then
         newen <= '0';
         cond <= '0';
         begcond <= '0';
         endcond <= '0';
      elsif rising_edge (CLK) then
         if (t = c) then
            cond <= '1';
            if (ec = et) then
               endcond <= '1';
               begcond <= '0';
            else
               begcond <= endcond;
               endcond <= '0';
            end if;
            if (endcond = '1') and (en = '1') then
               newen <= '1';  -- only start at beginning of cycle
            end if;
         else
            cond <= '0';
         end if;
      end if;
   end process;

   process (CLK, en, olden) is
   begin
      if rising_edge (CLK) then
         olden <= olden(0) & en;
      end if;
   end process;

   GO <= newen and cond;
   FIRST <= begcond;
   LAST <= endcond;
   FLUSH <= olden(1) and not olden(0);

end Structure;
--------------------------------------------------------------------------------
